/* HTTaP.c
   (C) 2011,2017,2020 Yann Guidon whygee@f-cpu.org
   The source code of this project is released under the AGPLv3 or later,
   see https://www.gnu.org/licenses/agpl-3.0.en.html

version 20170327: split from serv_11_CORS.c, a lot of renaming and translations
version 20170331: timeout and connection persistence work,
                  HTTaP and file serving appeared
version 20200415: removed IPv6, added more checks & some portability,
                  simplified some parts
version 20200416: solved a lot of little problems, works better than yesterday.
version 20200425: completely restructured around poll()
version 20200427: keepalive_counter removed
version 20200427: reworked...
version 20200501: missing close: finally solved
version 20200504: added itos()
version 20200505: added session id

TODO : replace strncpy with memcpy ?

This file is included by HTTaP.h, where you can configure the parameters.
 */

#define HTTAP_TCP_BACKLOG 0 /* parameter for listen(2) */

#include <poll.h>

struct pollfd pollfd_array[2]={
  { SOCKET_ERROR, POLLIN, 0 }, // server socket, "readable" when accept() is possible
  { SOCKET_ERROR, POLLIN, 0 }  // Client socket, "writeable" only when open
};

/* persistent variables */
//int sock_server, sock_client; moved/aliased directly to the above array:
#define sock_server pollfd_array[0].fd
#define sock_client pollfd_array[1].fd

int nb_pollfd;  // =1 when waiting, =2 when client connection open

struct sockaddr_in sockaddr_client;
socklen_t size_client = sizeof(sockaddr_client);

#ifdef HTTAP_SERVE_FILES
struct stat stat_buf;
int root_page_len; // size of the "index.html" string
char *root_page;  // string to "index.html"
uid_t prog_uid;
#endif

char HTTaP_buffer[HTTAP_BUFFER_LEN];  // shared receive and emit buffer.

#ifdef HTTAP_VERBOSE
#define HEADER_BUFFER_LEN (1024) // just to display the header request safely
char header_buffer[HEADER_BUFFER_LEN];
#endif

///////////////////////////////////////////////////////////////

  // this is called just after a call to clock_gettime()
  // so timestamp is updated. Base32 is not hard to code
  // and 8 chars provide 40 bits of entropy.
void update_httap_session() {
  uint64_t t=timestamp.tv_sec;
  int i=0;
  char c;
  do {
    c = t & 31;
    t = t >> 5;
    if (c < 10)
      c = c+'0';
    else
      c = (c-10)+'a';
    HTTaP_header[HTTAP_SESSION_OFFSET+(i++)]=c;
  } while (i < HTTAP_SESSION_LENGTH);

#ifdef HTTAP_VERBOSE
  HTTaP_header[HTTAP_SESSION_OFFSET+HTTAP_SESSION_LENGTH]=0xA;
  HTTaP_header[HTTAP_SESSION_OFFSET+HTTAP_SESSION_LENGTH+1]=0;
  printf(HTTaP_header+HTTAP_SESSION_ORIGIN);
  HTTaP_header[HTTAP_SESSION_OFFSET+HTTAP_SESSION_LENGTH]=0xD;
  HTTaP_header[HTTAP_SESSION_OFFSET+HTTAP_SESSION_LENGTH+1]=0xA;
#endif
}

///////////////////////////////////////////////////////////////

int close_client() {
  nb_pollfd=1;
  closesocket(sock_client);
  sock_client = INVALID_SOCKET;
  VERBOSE_PRINTF("\n###################################\n\n");
  return 2; // force change of poll state
}

///////////////////////////////////////////////////////////////

void HTTaP_init() {
  int i;
  char *env_port, *b;
  struct addrinfo hints, chosen_ai;

#ifdef HTTAP_SERVE_FILES
  char *env_user, *env_group;
  struct passwd *pswd=NULL;
  struct group *grp=NULL;
  char *static_path;
#endif

    /* Configuration */

  /* Port number */
  env_port = my_getenv(HTTAP_TCPPORT_NAME);
  if (!env_port) {
    env_port=HTTAP_TCPPORT;
    i = atol(env_port);
    if ((i < 2) || (i > 65535))
      show_exit("$" HTTAP_TCPPORT_NAME " must be between 1 and 65535\n", 0);
  }

  /* How long can a connection remain open ? */
  keepalive_seconds = HTTAP_KEEPALIVE;
  b = my_getenv(HTTAP_KEEPALIVE_NAME);
  if (b) {
    keepalive_seconds = atol(b);
    if ((keepalive_seconds < 2)
     || (keepalive_seconds > 200))
      show_exit("$" HTTAP_KEEPALIVE_NAME " must be between 2 and 200\n", 0);
  }
  VERBOSE_PRINTF("Port: %s, Keepalive: %ds\n", env_port, keepalive_seconds);

#ifdef HTTAP_SERVE_FILES
  /* Path to the directory that contains the static files */
  static_path = my_getenv(HTTAP_STATICPATH_NAME);
  if (!static_path)
    static_path = "."; /* The current working directory */

  /* Filename of the root URI (by default index.html) */
  root_page = my_getenv(HTTAP_ROOTPAGE_NAME);
  if (!root_page)
    root_page = HTTAP_ROOTPAGE;
  VERBOSE_PRINTF("Path: %s  Root page: %s\n", static_path, root_page);
  root_page_len=strlen(root_page);

  /* Check if we change the user and group */
  env_user = my_getenv(HTTAP_USER_NAME);
  if (env_user) {
    errno = 0;
    pswd = getpwnam(env_user);
    if (!pswd)
      show_exit("$" HTTAP_USER_NAME " is invalid ",0);

    VERBOSE_PRINTF("User: %s", env_user);
    prog_uid = pswd->pw_uid;

    env_group = my_getenv(HTTAP_GROUP_NAME);
    if (env_group) {
      errno = 0;
      grp = getgrnam(env_group);
      if (!grp)
        show_exit("$" HTTAP_GROUP_NAME "is invalid ", 0);

      VERBOSE_PRINTF(", Group: %s", env_group);
    }
  }
  else
    prog_uid = getuid();
  VERBOSE_PRINTF("\n");

  /* Check the path to access the static files */
  if (stat(static_path, &stat_buf)
   || !S_ISDIR(stat_buf.st_mode))
    show_exit("$" HTTAP_STATICPATH_NAME " : invalid path / directory name", 0);
  if (stat_buf.st_uid != prog_uid)
    show_exit("The program's user is not owner of $" HTTAP_STATICPATH_NAME "\n", 1);
  if (chdir(static_path))
    show_exit("Failed to change directory to $" HTTAP_STATICPATH_NAME , 1);
  if (chroot(static_path))
    perror("Warning: chroot() failure (are you root ?) ");

  /* Check the default root file */
  if (lstat(root_page, &stat_buf)
   || !S_ISREG(stat_buf.st_mode))
    show_exit("$" HTTAP_ROOTPAGE_NAME " : invalid root file", 1);
  if (stat_buf.st_uid != prog_uid)
    show_exit("The user doesn't own the root page $" HTTAP_ROOTPAGE_NAME "\n", 0);

  /* change the group THEN the user, after chroot() */
  if (pswd){ /* $HTTAP_USER exists */
    if (grp) {
      if (setgid(grp->gr_gid))
        show_exit("Failed to change running group ", 1);
    }
    if (setuid(pswd->pw_uid))
      show_exit("Failed to change running user ", 1);
  }
#endif

/*
   Open the socket
*/

  /* Describes the type of socket we want to open */
  memset(&hints, 0, sizeof(hints)); /* cleanup */
  hints.ai_family   = AF_INET;      /* IPv4 */
  hints.ai_socktype = SOCK_STREAM;  /* TCP */
  hints.ai_flags    = AI_PASSIVE;   /* server mode */

  if ((sock_server = create_IPv4_socket(NULL,
        env_port, &chosen_ai, &hints) ) == SOCKET_ERROR)
    show_exit("Failed to create the server.", 0);
  VERBOSE_PRINTF("\n");

#ifdef SO_REUSEADDR
  if (setsockopt(sock_server, SOL_SOCKET, SO_REUSEADDR, &const_one, sizeof(const_one)) == SOCKET_ERROR)
    net_perror("setsockopt(SO_REUSEADDR)");
#endif

#ifdef SO_REUSEPORT
  if (setsockopt(sock_server, SOL_SOCKET, SO_REUSEPORT, &const_one, sizeof(const_one)) == SOCKET_ERROR)
    net_perror("setsockopt(SO_REUSEPORT)");
#endif

  /* Connect the socket */
  if (bind(sock_server, chosen_ai.ai_addr, chosen_ai.ai_addrlen) == SOCKET_ERROR)
    net_exit("Failed to bind() the server");

  /* Listen on the socket */
  if (listen(sock_server, HTTAP_TCP_BACKLOG) == SOCKET_ERROR)
    net_exit("Failed to listen()");

  nb_pollfd=1;
  VERBOSE_PRINTF("Server socket ready\n");
}


 /**************************************
       Examine the client's request
  **************************************/

int HTTaP_process_packet(int recv_len) {
  int ret=0, // this is the return value:
        // 0 if nothing happens
        // 1 if something happened and socket is to be kept alive.
        // 2 if socket has been closed.
      len=sizeof(err400)-1; // size of the reply packet
  char *b=err400, // setup the default error.
       *fname;    // pointer to parse the URI
#ifdef HTTAP_SERVE_FILES
  int readfd=-1, headfd=-1, i, j, k, l, offset;
  char c, d;
#endif

#ifdef HTTAP_VERBOSE
  int n, max = HEADER_BUFFER_LEN-5; // count the 4 trailing chars
  if (recv_len < HEADER_BUFFER_LEN-5)
    max = recv_len;

  printf("received %d bytes\n---------------------------\n", recv_len);
  k = 0;
  // safe-copy the first 4 lines of header/request to header_buffer,
  // max size is HEADER_BUFFER_LEN (give or take some terminator)
  for (n=0; (k<8) && (n<max); n++) {
    c=HTTaP_buffer[n];
    if ((c<' ') || (c >= 127)) {
      if ((c!=0xD) && (c!=0xA))
        c='!';
      else
        k++;
    }
    header_buffer[n]=c;
  }
  header_buffer[n+0]='\n';
  header_buffer[n+1]='*';
  header_buffer[n+2]='\n';
  header_buffer[n+3]=0;
  printf(header_buffer);
#endif

  // First, check the method :
  if ((recv_len >= GET_HEADER_LEN)
   && (memcmp(GET_header, HTTaP_buffer, GET_HEADER_LEN) == 0)) {

    fname = &HTTaP_buffer[GET_HEADER_LEN]; // now point beyond the '/'
    recv_len -= GET_HEADER_LEN;

    /*
       HTTaP
     */

    if (( recv_len > 0) && (fname[0]=='?')) {
      fname++; // now point beyond the '?'
      VERBOSE_PRINTF("HTTaP : ");

      // check the Default root
      if ((recv_len == 1)
       ||((recv_len >  1)
         && ( (*fname==' ')
           || (*fname==0xD)
           || (*fname==0xA) ))) {
        VERBOSE_PRINTF("root\n");

        // assemble the header
        memcpy(HTTaP_buffer, HTTaP_header,
          (len = sizeof(HTTaP_header)-1));
        // add the length (but remove the 4 DADA chars)
        len += itos( sizeof(HTTaP_json_root)-(1+4),
                HTTaP_buffer+(sizeof(HTTaP_header)-1));
        // end with the payload
        memcpy(HTTaP_buffer+len, HTTaP_json_root, sizeof(HTTaP_json_root)-1);
        len += sizeof(HTTaP_json_root)-1;

        if (len >= HTTAP_BUFFER_LEN) { // this should never happen, right ?
          VERBOSE_PRINTF("Output buffer saturated ?\n");
          len=sizeof(err500)-1;
          b=err500;
        }
        else {
          b=HTTaP_buffer;
          ret=1;
        }
      } // root
      else {
        if ((recv_len >= PING_HEADER_LEN)
         // there must be a way to check for a reasonable max. length here
         && (memcmp(PING_header, fname, PING_HEADER_LEN)==0 )) {
           // The test is weak because *anything* is
           // allowed after the keyword to help prevent caching.
           // TODO make sure to flush the input buffers...

          // yes, there can be some time between the last call to get_time_diff()
          // and the return of poll() so we must measure it again.
          if (PollState == H_BLOCKING)
            get_time_diff();

          // check bounds to fit the number in 4 digits only
          // (which should never have to be done anyway)
          if (H_timediff > 9999)
              H_timediff = 9999;  // which is not possible because init() limits to 200s
          else
            if (H_timediff <  0)  // this should not even happen, ever...
                H_timediff = -1;
          n=itos(H_timediff, pong200+OFFSET_PONG_VAL);
          // pad right with spaces :
          while (n<4) {
            pong200[OFFSET_PONG_VAL+n]=' ';
            n++;
          }

          VERBOSE_PRINTF(pong200+6);

          // assemble the header : I suppose these could be merged ?
          // but C can't ensure these constants are stored in consecutive places :-(
          memcpy(HTTaP_buffer, HTTaP_header, sizeof(HTTaP_header)-1);
          memcpy(HTTaP_buffer+(sizeof(HTTaP_header)-1), pong200, sizeof(pong200)-2);

          ret=1; // valid reply.
          len= (sizeof(HTTaP_header)-1) + (sizeof(pong200)-2);
          b=HTTaP_buffer;
        }
        else {
          // insert your own HTTaP code here :-)
          // don't forget to set these variables if the request ends well
          // b=HTTaP_buffer; // or somewhere else
          // len = ?
          // ret 1; ?

#ifdef HTTAP_PARSE
          ret = HTTAP_PARSE( // return 0 : close, return 1: don't close
            fname,     // pointer to the request
            recv_len,  // byte count of the request
            1,         // 1 for GET, 2 for POST
            b,         // pointer to the send buffer
            &len);     // length of the buffer
#else
          // for now, just complain (in case there is a fbclid suffix)
          VERBOSE_PRINTF(" key not found ");
          len=sizeof(err400key)-1;
          b=err400key;
#endif
          // goto send_short_block;
        }
      }
      // now go to the send() section
    } // End of HTTaP domain

#ifdef HTTAP_SERVE_FILES
      /*
          The static file server
       */

    else {
      if ((recv_len == 0)
       ||((recv_len >  0)
         && ( (*fname==' ')
           || (*fname=='#')
           || (*fname=='?')
           || (*fname==0xD)       // These ones were added
           || (*fname==0xA) ))) { // for manual tests with telnet
        VERBOSE_PRINTF("got root\n");
        // Substitute the blank URI with the index.html page.
        memcpy(fname, root_page, root_page_len+1);
        recv_len=root_page_len+1;
      }

      l=0;
      d=0;
      // scan the URI and clean it up
Scan_URI_Loop:
      c=fname[l];
      switch(c){
        case 0xD:
        case 0xA:
        case '#':
        case '?':
        case ' ':
        case 0:
          goto endscanURI;

        case '.':
          // prevent going to the root directory
          if (d=='.') goto send_short_block;
          break;
        case '/':
         // prevent going to the root directory
          if (d==0) goto send_short_block;
          break;
        case '-':  // only allowed characters
        case '_':  // (no '%' funky business)
          break;

        default:
          if (c< '0') goto send_short_block;
          if (c<='9') break;
          if (c< 'A') goto send_short_block;
          if (c<='Z') break;
          if((c< 'a')
          || (c> 'z')) goto send_short_block;
        // fold
      }
      d=c; // remember for the next iteration
      if (++l > MAX_URI_LEN-(GET_HEADER_LEN+1)) {
        VERBOSE_PRINTF("URI size overflow: skip everything\n");
        goto send_short_block;
      }
      if (l < recv_len)
        goto Scan_URI_Loop; /* one more loop to process the next character */

endscanURI:
      fname[l]=0; /* coming from "case ' ':", terminate the string */
      VERBOSE_PRINTF("trying to read file: %s\n", fname);

      if (lstat(fname, &stat_buf) == -1) {
        perror("stat error : ");
        switch (errno) {
          case EACCES:
            VERBOSE_PRINTF("403: EACCES\n");
            len=sizeof(err403)-1;
            b=err404;
            break;

          case ENOTDIR:
          case ENOENT:
            VERBOSE_PRINTF("404: ENOTDIR/ENOENT\n");
            len=sizeof(err404)-1;
            b=err404;
            break;

          default:
            VERBOSE_PRINTF("500: others\n");
            len=sizeof(err500)-1;
            b=err500;
        }
        goto send_short_block;
      }

      if (!S_ISREG(stat_buf.st_mode)
       || stat_buf.st_uid != prog_uid) {
        VERBOSE_PRINTF("403: bad UID or file type ?\n");
        len=sizeof(err403)-1;
        b=err403;
        goto send_short_block;
      }

      /* and now, open the actual data file */
      readfd=open(HTTaP_buffer+5, O_RDONLY);
      if (readfd == -1) {
        perror("can't open data file ! ");
        len=sizeof(err500)-1;
        b=err500;
        goto send_short_block;
      }

      /* locate the file that contains the custom headers */
      HTTaP_buffer[1]='.';
      HTTaP_buffer[2]='.'; // prepend "..." to the path
      HTTaP_buffer[3]='.';
      headfd=open(HTTaP_buffer+1, O_RDONLY);

  //   At this point : HTTaP_buffer is not read anymore.
  //   We can now overwrite it and build the response message there.

      // first: copy the 200 OK message
      memcpy(HTTaP_buffer, ok200, sizeof(ok200)-1);
      offset=sizeof(ok200)-1;

      // now check the extra headers :
      if (headfd == -1) {
        VERBOSE_PRINTF("Can't access the extra header file\n");
      }
      else {
        i=read(headfd, &(HTTaP_buffer[sizeof(ok200)-1]),
            (HTTAP_BUFFER_LEN-42)-sizeof(ok200));
              // That string is copy-pasted a few times in this block.
              // 42 is a nice number...
              // ContentLength[] shoudn't be longer than 30,
              // and we need some space (10 bytes ?) to read the file's first bytes.
        if (i == -1) {
          VERBOSE_PRINTF("Can't read the extra header file\n");
        }
        else {
          if (i>= ((HTTAP_BUFFER_LEN-42)-sizeof(ok200))) {
            printf("\nWARNING ! extra header file: file too long!\n");
          }
          else {
            VERBOSE_PRINTF("Extra Header: %d bytes\n", i);
            HTTaP_buffer[(sizeof(ok200)-1)+i]=0; // zero termination
            VERBOSE_PRINTF(&(HTTaP_buffer[sizeof(ok200)-1]));
            offset+=i;
          }
        }
        close(headfd);
      }

      /* Assemble the last header : Content-Length: */
      // offset+=sprintf(HTTaP_buffer+offset, ContentLength, (int)stat_buf.st_size);
      memcpy(HTTaP_buffer+offset, ContentLength, sizeof(ContentLength)-1);
      offset+=sizeof(ContentLength)-1;
      offset+=itos((int)stat_buf.st_size, HTTaP_buffer+offset);
      HTTaP_buffer[offset  ] = 0xD;
      HTTaP_buffer[offset+1] = 0xA;
      HTTaP_buffer[offset+2] = 0xD;
      HTTaP_buffer[offset+3] = 0xA;
      offset+=4;

      // this should be prevented by design...
       // but what keeps users from allocating very small buffers ?
      if (offset >= HTTAP_BUFFER_LEN-10) {
        close(readfd);
        VERBOSE_PRINTF("Header overflow, err500\n");
        len=sizeof(err500)-1;
        b=err500; // Check overflow, unlikely but you never know
        goto send_short_block;
      }

      VERBOSE_PRINTF("sending %d bytes header + %d bytes payload : ",
           offset, (int)stat_buf.st_size);

// man 2 sendfile says : "not portable" :-/
// so here is the slow version:

      do {
        /* Read and appends the file's contents after the header */
        j=HTTAP_BUFFER_LEN-offset;
        i=read(readfd, HTTaP_buffer+offset, j);
        if (i == -1) {
          close(readfd);
          perror("can't read the data file ! ");
          if (offset) {  // the header can be changed
            len=sizeof(err500)-1;
            b=err500;
            goto send_short_block;
          }
          else
            goto close_sock_client; // simply abort
            // at this point because the header is already sent
        }
        k=i+offset;
        offset=0;

        if (k) { // don't call in the unlikely event that nothing has been read
          /* send the header and the available data */
          if (send(sock_client, HTTaP_buffer, k, 0) != k) {
            close(readfd);
            net_perror("Failed to send the reply");
            goto close_sock_client;
          }
          VERBOSE_PRINTF("%d ", k);
        }
      } while (i==j);
      close(readfd);
      VERBOSE_PRINTF("\n");
      return 1; // do not fold to sending the
                // small packet, return successfully !
    }
#endif  //  HTTAP_SERVE_FILES

  } // end of method GET
  else {
    if ((recv_len >= POST_HEADER_LEN)
     && (memcmp(POST_header, HTTaP_buffer, POST_HEADER_LEN) == 0)) {
      VERBOSE_PRINTF("POST !\n");
      fname = &HTTaP_buffer[POST_HEADER_LEN]; // now point beyond the '/'
      recv_len -= POST_HEADER_LEN;

//
// POST stuff comes here.
//

#ifdef HTTAP_PARSE

      if (( recv_len > 0) && (fname[0]=='?')) {
        recv_len--;
        fname++; // now point beyond the '?'
        VERBOSE_PRINTF("HTTaP : ");

        ret = HTTAP_PARSE( // return 0 : close, return 1: don't close
          fname,     // pointer to the request
          recv_len,  // byte count of the request
          2,         // 1 for GET, 2 for POST
          b,         // pointer to the send buffer
          &len);     // length of the buffer
      // goto send_short_block;
      }
      else
#endif

      // Loopback ? not yet implemented...

    } // end of method POST
    else {
      VERBOSE_PRINTF("405: not GET or POST :-(\n");
      b=err405; // Method not allowed.
      len=sizeof(err405)-1;
    }
  }

  // Send a simple reply :
send_short_block:
  if (len >= HTTAP_BUFFER_LEN) // Should never happen...
    show_exit("Can't send this packet, too long !\n", 0);
  // Now b and len must be valid here.
  if (send(sock_client, b, len, 0) != len) {
    net_perror("Failed to send the reply");
    goto close_sock_client;
  }
  VERBOSE_PRINTF("Sent %d bytes of reply\n", len);
  if (ret==0) {
close_sock_client:
    return close_client();
  }

  return 1;
}


/////////////////////////////////////////////////////////////////////

int HTTaP_server(int timeout) {
   // return value :
   // 0 is nothing happens
   // 1 if something happened and socket is to be kept alive.
   // 2 if socket has been close.
  short ev;
  SOCKET_TYPE sock;
  int r,
    recv_len, // size of the received packet
    p=poll( &pollfd_array[0], nb_pollfd, timeout?0:H_poll_timeout);

  if (p==SOCKET_ERROR) {
    if (LastSockError() != INTR)
      net_exit("\npoll() error ");
    // and now jump to the end of the function to check the timeout flag
  }
  else {  // did something happen ?

    // FIRST Check the client but only if it is open :
    if (nb_pollfd > 1) {
      ev = pollfd_array[1].revents;
      if (ev) {
        if (ev & ~POLLIN)
          net_exit("\npoll() problem on client socket \n");

        // finally, something to actually do, let's read that message (non-blocking way)
        recv_len=recv(sock_client, HTTaP_buffer, HTTAP_BUFFER_LEN, MSG_DONTWAIT);
        if (recv_len==SOCKET_ERROR) {
          r=LastSockError(x);
          if ((r != EWOULDBLOCK) && (r != EAGAIN)) {
            net_perror("Failed to read the client's request");
            return close_client();
          }
          VERBOSE_PRINTF(" recv() <= EWOULDBLOCK || EAGAIN ");
          // don't return yet, check the server socket, just in case ?
        }
        else {
          if (recv_len > 0)
            return HTTaP_process_packet(recv_len);

          // recv_len == 0:
          return close_client();  // socket closed by the peer, close the local socket
        }
      }
    } // End of processing client socket

    // Check if the server socket has a new request:
    ev = pollfd_array[0].revents;
    if (ev) {
      if (ev & ~POLLIN)
        net_exit("\npoll() problem on server socket \n");

      // ok : it's a connection request if you can read on that socket.
      if ((sock = accept(sock_server,
       (struct sockaddr *) &sockaddr_client, &size_client)) == SOCKET_ERROR)
        net_exit("accept(): Failure to wait for a network connection");

      // reject the new connection if we're already busy:
      if (sock_client != SOCKET_ERROR) {
        VERBOSE_PRINTF("Closing one spurious client\n");
        closesocket(sock);
        // don't return yet, check the timeout
      }
      else {
        // register our new client
        sock_client = sock;
        nb_pollfd=2;

        /* Disable Nagle : it's very good to have but not a critical feature */
        if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &const_one, sizeof(const_one)) < 0)
          net_perror("Warning : setsockopt(TCP_NODELAY) / Can't disable Nagle ");

#ifdef SHOW_CLIENT
        printf("* Connected to %s:%d\n",
             inet_ntoa(sockaddr_client.sin_addr), ntohs(sockaddr_client.sin_port));
#endif

        // update_httap_session(); is moved to HTTaP_manager()

        return 1; // yay ! something happened !
      } // new sock registered
    } // new socket accept() ?
  } // poll() had something new for us

  if (timeout)
    return close_client();

  return 0; // really nothing happened.
}
